/***    MRCI.H - Microsoft Real-Time Compression Interface definitions
*
*       MRCI version 1.00.08  02-Apr-1993
*/

#define intMRCI         (0x2F)      /* MRCI interrupt number                */
#define mrciDETECT      (0x4A12)    /* intMRCI decection function           */

#define intMRCIROM      (0x1A)      /* ROM MRCI server interrupt number     */
#define mrciDETECTROM   (0xB001)    /* intMRCIROM detection function        */


/***    mcXXXX - flag values passed to MRCI operations
*
*       MRCCompress and MRCDecompress take a flag to indicate whether the
*       client is a *system* component (and hence may call with InDOS
*       set), or an *application*.
*
*       If a system client calls MRCCompress/MRCDecompress with mcSYSTEM,
*       then it is the clients responsibility to ensure that one of the
*       following is true:
*           1)  InDOS is set
*           2)  The call is made at interrupt time.
*
*       FAILING TO FOLLOW THE ABOVE RULE WILL LIKELY RESULT IN A SYSTEM
*       HANG AND LOSS OF USER DATA.
*/

#define mcAPPLICATION   (0)         /* Client is an application             */
#define mcSYSTEM        (1)         /* Client is a file system driver       */


/***    micapXXXXX - bit flags for MRCINFO.mi_flCapabilities
*
*       These define both the capabilities of the Server, and also double
*       as *operation* codes passed to the mi_pfnOperate entry point in
*       the server.
*/
#define micapNONE        (0x0000)   /* No capabilities                      */
#define micapSTANDARD    (0x0001)   /* Standard compress                    */
#define micapDECOMPRESS  (0x0002)   /* Standard decompress                  */
#define micapRESERVED1   (0x0004)   /* RESERVED1                            */
#define micapRESERVEDM   (0x0008)   /* RESERVEDM                            */
#define micapRESERVED2   (0x0010)   /* RESERVED2                            */
#define micapINCDECOMP   (0x0020)   /* Incremental Decompress               */

                       /* Remaining bits (6..14) are RESERVED and must be 0 */

#define micapREADONLY    (0x8000)   /* MRCINFO structure is read-only       */
#define micapDEINSTALL   (0xFFFF)   /* Deinstall server                     */


/***    MRCINFO - MRC Information data structure
*
*       A pointer to an MRCINFO structure is returned from MRCQuery, and sent
*       on MRCNotifyLoad.  This structure contains information on the MRCI
*       server and its capabilities.
*/

typedef struct {
    unsigned long   mi_lVendor;         /* A 4-byte vendor ID.
                                           Microsoft's vendor ID is "MSFT". */

    unsigned short  mi_wVendorVersion;  /* Version number of the MRC server.
                                           High byte is major number, low byte
                                           byte is minor.  EXAMPLES: v3.20 =
                                           0314h, v10.01 = 0A01h            */

    unsigned short  mi_wMRCIVersion;    /* Version number of the MRCI
                                           supported by this server.        */

    int      (_far *mi_pfnOperate)();   /* Far pointer of the server
                                           compression entry point.         */

    unsigned short  mi_flCapability;    /* Bit field of server capabilities.
                                           See micapXXX for bit definitions.*/

    unsigned short  mi_flHWAssist;      /* Bit field of hardware assisted
                                           server capabilities.  One-to-one
                                           correspondance with mi_flCapability
                                           bits.  A bit set in this field
                                           indicates the corresponding
                                           capability is hardware assisted
                                           (fast). */

    unsigned short  mi_cbMax;           /* Maximum number of bytes that the
                                           compression services provider can
                                           compress or decompress.  Requests
                                           to compress or decompress buffers
                                           in excess of this length will fail.
                                           All MRCI servers are to support at
                                           least 8192 byte (8Kb) blocks.    */

} MRCINFO, _far *LPMRCINFO;


/***    MRCREQUEST - MRC compress/decompress Request packet
*
*       This structure is used to pass parameters to the server for
*       compress/decompress operations.
*
*       General Notes
*       -------------
*       (1) <mr_pbSrc,mr_cbSrc> and <mr_pbDst,mr_cbDst> MUST NOT
*           OVERLAP!
*
*       (2) The safest practice is for mr_cbSrc and mr_cbDst to be
*           identical.
*
*       Details on Structure Members
*       ----------------------------
*       mr_pbSrc
*           This points to the *source* buffer.
*
*           On a *compress* operation, the contents of this buffer are
*           *uncompressed* data.
*
*           On a *decompress* operation, the contents of this buffer are
*           *compressed* data.
*
*           On a *incremental decompress* operation, this field points to
*           the next section of compressed data to be uncompressed.  The
*           server updates the offset portion of this address after each
*           incremental decompress call and the application should not modify
*           this address between incremental decompress calls on the same block
*           of compressed data.
*
*       mr_cbSrc
*           This is the size of the *source* buffer.
*
*           On a *compress* operation, this is the amount of data to
*           be compressed.
*
*           For a *decompress* operation, this value is ignored.  The amount
*           of data to be decompressed is specified by the mr_cbDst parameter,
*           described below.
*
*       mr_RESERVED
*           RESERVED for future use.  Should be 0.
*
*       mr_pbDst
*           This points to the *destination* buffer.
*
*           On a *compress* operation, this buffer receives the *compressed*
*           result of the operation.
*
*           On a *decompress* operation, this buffer receives the
*           *uncompressed* result of the operation.
*
*           On a *incremental decompress* operation, this field points to
*           the next location in the destination buffer where uncompressed
*           data is to be stored.  The server updates the offset portion of
*           this address after each incremental decompress call, and the
*           application should not modify this address between incremental
*           decompress calls on the same block of compressed data.
*
*       mr_cbDst
*           On INPUT, for a *compress* operation, this is the size of the
*           *destination* buffer. If the compressed data would overflow
*           this buffer length, then the operation fails and the server
*           returns the error MRCI_ERROR_BUFFER_OVERFLOW.
*
*           On INPUT, for a *decompress* operation, this must be EXACTLY the
*           number of bytes that will be decompressed, as the MRCI server
*           uses this information to determine when to stop decompressing.
*
*           On INPUT, for an *incremental decompress* operation, this is
*           the number of bytes that should uncompressed at this time.  This
*           will typically be less than the original uncompressed size of the
*           compressed block.  A single compressed block can be uncompressed in
*           steps by making multiple incremental decompress calls for smaller
*           sized blocks.  However, to incrementally decompress an entire
*           compressed block requires that the sum of the individual mr_cbDst
*           counts be EXACTLY the number of bytes that were originally
*           compressed so the MRCI server can determine when to stop
*           decompressing.
*
*
*           On OUTPUT, the Server updates this field with the actual size
*           of the resulting compressed/uncompressed data.
*
*       mr_cbChunk
*           This is information that the Server compress routines can use to
*           "early out" of the compression as early as possible.
*
*           Valid values are 1 (client is interested in savings as small as
*           1 byte) to 32767.  DblSpace passes 512, and Flash File System
*           passes 1.
*
*           This field is most easily explained by giving an example:
*
*           Example:
*               DblSpace does space allocation in chunks of 512 bytes (the
*               common sector size on a disk).
*
*               The compression server can use this information for two
*               optimizations:
*
*               (1) If the Server cannot compress the uncompressed data
*                   enough to save at least 512 bytes, then the data is
*                   *incompressible* as far as DblSpace is concerned, even
*                   if it could be compressed to save fewer than 512 bytes.
*
*               (2) While compressing, if the Server gets to a point where
*                   the remaining uncompressed data is of such a length that
*                   it can be encoded simply (without table lookups, etc.)
*                   and not cross a 512 byte boundary, then the Server can
*                   do the simple encoding.
*
*           It is likely that these optimizations will be hard to perform
*           quickly in software, but it is possible that hardware can do
*           these optimizations without any performance loss.
*
*       mr_dwIncDecomp
*           NOTE: This is used for Incremental Decompression only.
*
*           For the first *incremental decompression* call on a compressed
*           block, this value must be set to zero.  Upon return, the field
*           will contain state information for use on the next incremental
*           decompress call.  This value must not be modified between
*           subsequent incremental decompress calls on the same compressed
*           block.
*/

typedef struct {

    void _far      *mr_pbSrc;       /* Pointer to source buffer             */
    unsigned short  mr_cbSrc;       /* Size of source buffer, in bytes
                                       (0 = 64K)                            */
    unsigned short  mr_RESERVED;    /* Reserved.  Should be 0               */
    void _far      *mr_pbDst;       /* Pointer to destination buffer        */
    unsigned short  mr_cbDst;       /* Size of destination buffer, in bytes
                                       (0 = 64K)                            */
    unsigned short  mr_cbChunk;     /* Client compressed data storage chunk
                                       size (see above!)                    */
    unsigned long   mr_dwIncDecomp; /* Incremental decompression state data */

} MRCREQUEST;


/***    MRCI_ERROR_XXX definitions
*
*       Error codes returned from MRCIOperate
*
*/

#define MRCI_ERROR_NONE             (0)     /* No error                     */
#define MRCI_ERROR_NOT_SUPPORTED    (1)     /* Unsupported operation        */
#define MRCI_ERROR_BUSY             (2)     /* Server is busy               */
#define MRCI_ERROR_BUFFER_OVERFLOW  (3)     /* Destination buffer too small */
#define MRCI_ERROR_NOT_COMPRESSIBLE (4)     /* Data could not be compressed */
#define MRCI_ERROR_BAD_MRC_FORMAT   (5)     /* Compressed data format is bad*/


/***    sigOLD_CX, sigOLD_DX, sigNEW_CX, sigNEW_DX - MRCI Server detection
*
*       These values are used to verify that the response from issuing
*       intMRCI is coming from an MRCI server, and not some other piece
*       of code.
*
*       The *old* values are passed on the mrciQUERY call, and the server
*       must change CX/DX to the *new* values, so that the caller can
*       trust that the MRCI server was responding, and not some other
*       interrupt hook.
*
*       The Server uses this code sequence to transform CX/DX:
*
*           ;------------- entry: cx='ab'  dx='cd'
*
*           xchg    ch,cl       ; cx='ba'  dx='cd'
*           xchg    dh,dl       ; cx='ba'  dx='dc'
*           xchg    dx,cx       ; cx='dc'  dx='ba'
*/

#define sigOLD_CX   (0x4D52)    /* 'MR'                                 */
#define sigOLD_DX   (0x4349)    /* 'CI'                                 */

#define sigNEW_CX   (0x4943)    /* 'IC'                                 */
#define sigNEW_DX   (0x524D)    /* 'RM'                                 */
