;***    MRCI.INC - Microsoft Real-Time Compression Interface definitions
;
;       MRCI version 1.00.08  02-Apr-1993

intMRCI         equ       2Fh       ; MRCI interrupt number
mrciDETECT      equ    04A12h       ; intMRCI AX for detecting MRCI server

intMRCIROM      equ       1Ah       ; ROM MRCI interrupt number
mrciDETECTROM   equ    0B001h       ; intMRCIROM AX for detecting MRCI server


;***    mcXXXX - flag values passed to MRCI operations
;
;       MRCCompress and MRCDecompress take a flag to indicate whether the
;       client is a *system* component (and hence may call with InDOS
;       set), or an *application*.
;
;       mcSYSTEM clients must ensure the following is true before calling
;       MRCI:
;           1)  The Windows Disk Critical Section is owned
;           2)  The InDOS flag is *set*
;
;       mcAPPLICATION clients must ensure the following is true before calling
;       MRCI:
;           1)  The Windows Disk Critical Section is owned
;           2)  The InDOS flag is *clear*
;
;       FAILING TO FOLLOW THE ABOVE RULES WILL LIKELY RESULT IN A SYSTEM
;       HANG AND LOSS OF USER DATA.
;
mcAPPLICATION   equ     0   ; Client is an application
mcSYSTEM        equ     1   ; Client is a file system driver


;***    micapXXXXX - bit flags for MRCINFO.mi_flCapabilities
;
;       These define both the capabilities of the Server, and also double
;       as *operation* codes passed to the mi_pfnOperate entry point in
;       the server.
;                          111111
;                          5432109876543210
;                          ----------------
micapNONE           equ    0000000000000000b ; No capabilities
micapSTANDARD       equ    0000000000000001b ; Standard compress
micapDECOMPRESS     equ    0000000000000010b ; Standard compress
micapRESERVED1      equ    0000000000000100b ; RESERVED for future use
micapRESERVEDM      equ    0000000000001000b ; RESERVED for future use
micapRESERVED2      equ    0000000000010000b ; RESERVED for future use
micapINCDECOMP      equ    0000000000100000b ; Incremental Decompress
   ;
   ; Remaining bits (6..14) are RESERVED and must be 0
   ;
micapREADONLY       equ    1000000000000000b ; MRCINFO structure is read-only
micapDEINSTALL      equ    1111111111111111b ; Server deinstall service


;***    MRCINFO - MRC Information data structure
;
;       A pointer to an MRCINFO structure is returned from MRCQuery, and sent
;       on MRCNotifyLoad.  This structure contains information on the MRCI
;       server and its capabilities.

MRCINFO struc

mi_lVendor          dd  ?       ; A 4-byte vendor ID.
                                ; Microsoft's vendor ID is "MSFT".

mi_wVendorVersion   dw  ?       ; Version number of the MRC server.
                                ; High byte is major number, low byte is minor.
                                ; EXAMPLES: v3.20 = 0314h, v10.01 = 0A01h

mi_wMRCIVersion     dw  ?       ; Version number of the MRCI supported by
                                ; this server.

mi_pfnOperate       dd  ?       ; Far pointer of the server compression entry
                                ; point.
                                ; NOTE: Caller must ensure that the Windows
                                ;       critical section is held *before*
                                ;       calling this entry point!

mi_flCapability     dw  ?       ; Bit field of server capabilities
                                ; See micapXXX for bit definitions

mi_flHWAssist       dw  ?       ; Bit field of hardware assisted
                                ; server capabilities.  One-to-one
                                ; correspondance with mi_flCapability
                                ; bits.  A bit set in this field
                                ; indicates the corresponding
                                ; capability is hardware assisted.

mi_cbMax            dw  ?       ; Maximum number of bytes that the compression
                                ; services provider can compress or decompress.
                                ; Requests to compress or decompress buffers in
                                ; excess of this length will fail.
                                ; All MRCI servers are to support at least
                                ; 8192 byte (8Kb) blocks.
MRCINFO ends


;***    MRCREQUEST - MRC compress/decompress Request packet
;
;       This structure is used to pass parameters to the server for
;       compress/decompress operations.
;
;       General Notes
;       -------------
;       (1) <mr_pbSrc,mr_cbSrc> and <mr_pbDst,mr_cbDst> MUST NOT
;           OVERLAP!
;
;       (2) The safest practice is for mr_cbSrc and mr_cbDst to be
;           identical.
;
;       Details on Structure Members
;       ----------------------------
;       mr_pbSrc
;           This points to the *source* buffer.
;
;           On a *compress* operation, the contents of this buffer are
;           *uncompressed* data.
;
;           On a *decompress* operation, the contents of this buffer are
;           *compressed* data.
;
;           On a *incremental decompress* operation, this field points to
;           the next section of compressed data to be uncompressed.  The
;           server updates the offset portion of this address after each
;           incremental decompress call and the application should not modify
;           this address between incremental decompress calls on the same block
;           of compressed data.
;
;       mr_cbSrc
;           This is the size of the *source* buffer.
;
;           On a *compress* operation, this is the amount of data to
;           be compressed.
;
;           For a *decompress* operation, this value is ignored.  The amount
;           of data to be decompressed is specified by the mr_cbDst parameter,
;           described below.
;
;       mr_RESERVED
;           RESERVED for future use.  Should be 0.
;
;       mr_pbDst
;           This points to the *destination* buffer.
;
;           On a *compress* operation, this buffer receives the *compressed*
;           result of the operation.
;
;           On a *decompress* operation, this buffer receives the
;           *uncompressed* result of the operation.
;
;           On a *incremental decompress* operation, this field points to
;           the next location in the destination buffer where uncompressed
;           data is to be stored.  The server updates the offset portion of
;           this address after each incremental decompress call, and the
;           application should not modify this address between incremental
;           decompress calls on the same block of compressed data.
;
;       mr_cbDst
;           On INPUT, for a *compress* operation, this is the size of the
;           *destination* buffer. If the compressed data would overflow
;           this buffer length, then the operation fails and the server
;           returns the error MRCI_ERROR_BUFFER_OVERFLOW.
;
;           On INPUT, for a *decompress* operation, this must be EXACTLY the
;           number of bytes that will be decompressed, as the MRCI server
;           uses this information to determine when to stop decompressing.
;
;           On INPUT, for an *incremental decompress* operation, this is
;           the number of bytes that should uncompressed at this time.  This
;           will typically be less than the original uncompressed size of the
;           compressed block.  A single compressed block can be uncompressed in
;           steps by making multiple incremental decompress calls for smaller
;           sized blocks.  However, to incrementally decompress an entire
;           compressed block requires that the sum of the individual mr_cbDst
;           counts be EXACTLY the number of bytes that were originally
;           compressed so the MRCI server can determine when to stop
;           decompressing.
;
;
;           On OUTPUT, the Server updates this field with the actual size
;           of the resulting compressed/uncompressed data.
;
;       mr_cbChunk
;           This is information that the Server compress routines can use to
;           "early out" of the compression as early as possible.
;
;           Valid values are 1 (client is interested in savings as small as
;           1 byte) to 32767.  DblSpace passes 512, and Flash File System
;           passes 1.
;
;           This field is most easily explained by giving an example:
;
;           Example:
;               DblSpace does space allocation in chunks of 512 bytes (the
;               common sector size on a disk).
;
;               The compression server can use this information for two
;               optimizations:
;
;               (1) If the Server cannot compress the uncompressed data
;                   enough to save at least 512 bytes, then the data is
;                   *incompressible* as far as DblSpace is concerned, even
;                   if it could be compressed to save fewer than 512 bytes.
;
;               (2) While compressing, if the Server gets to a point where
;                   the remaining uncompressed data is of such a length that
;                   it can be encoded simply (without table lookups, etc.)
;                   and not cross a 512 byte boundary, then the Server can
;                   do the simple encoding.
;
;           It is likely that these optimizations will be hard to perform
;           quickly in software, but it is possible that hardware can do
;           these optimizations without any performance loss.
;
;       mr_dwIncDecomp
;           NOTE: This is used for Incremental Decompression only.
;
;           For the first *incremental decompression* call on a compressed
;           block, this value must be set to zero.  Upon return, the field
;           will contain state information for use on the next incremental
;           decompress call.  This value must not be modified between
;           subsequent incremental decompress calls on the same compressed
;           block.

MRCREQUEST  struc
mr_pbSrc        dd  ?   ; Pointer to source buffer
mr_cbSrc        dw  ?   ; Size of source buffer, in bytes
mr_RESERVED     dw  ?   ; RESERVED for future use.
mr_pbDst        dd  ?   ; Pointer to destination buffer
mr_cbDst        dw  ?   ; Size of destination buffer, in bytes
mr_cbChunk      dw  ?   ; Client compressed data storage chunk size (see above!)
mr_dwIncDecomp  dd  ?   ; Incremental Decompression state
MRCREQUEST  ends


;***    MRCI_ERROR_XXX definitions
;
;       Error codes returned from MRCIOperate
;
MRCI_ERROR_NONE             equ     0   ; No error
MRCI_ERROR_NOT_SUPPORTED    equ     1   ; Unsupported operation requested
MRCI_ERROR_BUSY             equ     2   ; Server is busy
MRCI_ERROR_BUFFER_OVERFLOW  equ     3   ; Destination buffer too small
MRCI_ERROR_NOT_COMPRESSIBLE equ     4   ; Data could not be compressed
MRCI_ERROR_BAD_MRC_FORMAT   equ     5   ; Compressed data format is bad

;***    DefineMRCQuery - Macro to generate MRCQuery function
;
;       Put this macro somewhere in your code segment.  It will define the
;       MRCQuery routine, which you can then call far.  See the MRCQuery
;       header below for documentation on its behavior.

DefineMRCQuery  macro


;***    sigOLD_CX, sigOLD_DX, sigNEW_CX, sigNEW_DX - MRCI Server detection
;
;       These values are used to verify that the response from issuing
;       intMRCI is coming from an MRCI server, and not some other piece
;       of code.
;
;       The *old* values are passed on the mrciQUERY call, and the server
;       must change CX/DX to the *new* values, so that the caller can
;       trust that the MRCI server was responding, and not some other
;       interrupt hook.
;
;       The Server uses this code sequence to transform CX/DX:
;
;           ;------------- entry: cx='ab'  dx='cd'
;
;           xchg    ch,cl       ; cx='ba'  dx='cd'
;           xchg    dh,dl       ; cx='ba'  dx='dc'
;           xchg    dx,cx       ; cx='dc'  dx='ba'
;
sigOLD_CX   equ 'MR'
sigOLD_DX   equ 'CI'

sigNEW_CX   equ 'IC'
sigNEW_DX   equ 'RM'


;***    MRCQuery - Detect presence of MRCI server, return MRCINFO
;
;       Detect presence of MRCI server safely, and if present return
;       pointer to the server's MRCINFO structure.  NOTE that we check
;       first for a RAM-based server, and then for a ROM-based server.
;
;       Entry
;           none
;
;       Exit-Success
;           ax = 0, MRCI server is present.
;           es:di -> MRCINFO structure
;
;
;       Exit-Failure
;           ax = 1, NO MRCI server is present.
;
;       Uses
;           ax,di,es,flags

MRCQuery  proc near
;*  Save caller's registers
        SaveReg <ax,bx,cx,dx,si,bp,ds>

;*  Check intMRCI vector before we issue the interrupt.
        xor     ax,ax                   ; Segment of interrupt vector table
        mov     ds,ax
        lds     si,ds:[intMRCI*4]       ; ds:si -> MRCI server

;*  Test if vector is plausible
        mov     ax,ds
        or      ax,ax                   ; Vector hooked?
        jz      mdr                     ;   NO, go make another check

;*  Call the server
        mov     ax,mrciDETECT           ; Function
        mov     cx,sigOLD_CX            ; Signatures for validation
        mov     dx,sigOLD_DX
        int     intMRCI                 ; Call server

        cmp     cx,sigNEW_CX            ; Signature match?
        jne     mdr                     ;   NO, go make another check

        cmp     dx,sigNEW_DX            ; Signature match?
        je      mdp                     ;   YES, have server

;*  Server not present, check for ROM based server
mdr:
        xor     ax,ax
        mov     ds,ax
        lds     si,ds:[intMRCIROM*4]    ; ds:si -> ROM MRCI server

;*  Test if vector is plausible
        mov     ax,ds
        or      ax,ax                   ; Vector hooked?
        jz      mde                     ;   NO, fail

;*  Call the server
        mov     ax,mrciDETECTROM        ; Function
        mov     cx,sigOLD_CX            ; Signatures for validation
        mov     dx,sigOLD_DX
        int     intMRCIROM              ; Call ROM server

        cmp     cx,sigNEW_CX            ; Signature match?
        jne     mde                     ;   NO, go make another check

        cmp     dx,sigNEW_DX            ; Signature match?
        jne     mde                     ;   No, fail

;*  Server is present
mdp:
        xor     ax,ax                   ; Indicate success
        jmp     short mdx               ; Go exit

;*  Set error
mde:    mov     ax,1                    ; Indicate failure

;*  Restore caller's registers and exit
mdx:    RestoreReg <ds,bp,si,dx,cx,bx,ax>
        ret
MRCQuery  endp

endm ;; DefineMRCQuery


;***    DefineMRCCompress - Macro to generate MRCCompress function
;
;       Put this macro somewhere in your code segment.  It will define the
;       MRCCompress routine, which you can then call far.  See the MRCCompress
;       header below for documentation on its behavior.

DefineMRCCompress   macro

;***    MRCCompress - Compress an uncompressed data buffer
;
;       Entry
;           ax = operation to perform:
;                   micapSTANDARD
;           cx = type of client:
;                   mcAPPLICATION - application
;                   mcSYSTEM      - file system client
;
;           ds:si -> MRCREQUEST structure
;               mr_pbSrc     - Pointer to uncompressed data buffer
;               mr_cbSrc     - Length of uncompressed data
;               mr_pbDst     - Pointer to compressed data buffer
;               mr_cbDst     - Length of compressed data buffer
;               mr_cbChunk   - Granularity of compressed data storage
;
;           es:bx -> MRCINFO structure returned by MRCQuery
;
;       Exit-Success
;           ax = 0, compress operation completed
;           ds:[si].mr_cbDst has length of compressed data in mr_pbDst.
;
;       Exit-Failure
;           Contents of mr_pbDst buffer and value of mr_cbDst are
;           undefined.
;           ax = non-zero error code:
;           MRCI_ERROR_NOT_SUPPORTED
;               Server does not support this operation.
;
;           MRCI_ERROR_BUSY
;               Server is busy with another operation.  Try again later.
;               NOTE: The most common case where this could occur is if an
;                     application calls the server while a disk cache
;                     (like SmartDrive) is writing its lazy-write queue at
;                     interrupt time to a compressed drive.  The application
;                     should try the operation again
;
;           MRCI_ERROR_BUFFER_OVERFLOW
;               The destination buffer size (mr_cbDst) was not large enough
;               to hold the compressed data.
;
;           MRCI_ERROR_NOT_COMPRESSIBLE
;               The data was not compressible, i.e., the size of the compressed
;               data would have been greater than (mr_cbDst - mr_cbChunk).
;
;       Uses
;           ax,flags

MRCCompress proc near
;*  Save caller's registers
        SaveReg <bx,cx,dx,bp,si,di,ds>

;*  Enter a Windows disk critical section.  This must be done to prevent
;   the MCRI server from being reentered if multiple VMs are making MRCI
;   calls under 386 Enhanced mode Windows.  The MRCI server sets the
;   InDOS flag, but that is not enough, as InDOS is a per-VM variable.

        push    ax                      ; NOTE: Use this exact sequence of
        mov     ax,8001h                ;       instructions because Windows
        int     2ah                     ;       expects it and will patch
        pop     ax                      ;       other code here

        call    dword ptr es:[bx].mi_pfnOperate ; Call Server

;*  Release the disk critical section.  The MRCI server is now available.

        push    ax                      ; NOTE: Use this exact sequence of
        mov     ax,8101h                ;       instructions because Windows
        int     2ah                     ;       expects it and will patch
        pop     ax                      ;       other code here

;*  Restore caller's registers and exit
        RestoreReg <ds,di,si,bp,dx,cx,bx>
        ret
MRCCompress endp

endm ;; DefineMRCCompress


;***    DefineMRCDecompress - Macro to generate MRCDeCompress function
;
;       Put this macro somewhere in your code segment.  It will define the
;       MRCDecompress routine, which you can then call far.
;       See the MRCDecompress header below for documentation on its behavior.

DefineMRCDecompress   macro


;***    MRCDecompress - Decompress an compressed data buffer
;
;       Entry
;           cx = type of client:
;                   mcAPPLICATION - application
;                   mcSYSTEM      - file system client
;
;           ds:si -> MRCREQUEST structure
;               mr_pbSrc   - Pointer to compressed data buffer
;               mr_cbSrc   - Length of compressed data
;               mr_pbDst   - Pointer to uncompressed data buffer
;               mr_cbDst   - Length of uncompressed data buffer
;
;           es:bx -> MRCINFO structure returned by MRCQuery
;
;       Exit-Success
;           ax = 0, decompress operation completed
;           ds:[si].mr_cbDst has length of uncompressed data in mr_pbDst.
;
;       Exit-Failure
;           Contents of mr_pbDst buffer and value of mr_cbDst are
;           undefined.
;           ax = non-zero error code:
;           MRCI_ERROR_NOT_SUPPORTED
;               Server does not support this operation.
;
;           MRCI_ERROR_BUSY
;               Server is busy with another operation.  Try again later.
;               NOTE: The most common case where this could occur is if an
;                     application calls the server while a disk cache
;                     (like SmartDrive) is reading ahead at interrupt time to
;                     a compressed drive.  The application should try the
;                     operation again.
;
;           MRCI_ERROR_BUFFER_OVERFLOW
;               The destination buffer size (mr_cbDst) was not large enough
;               to hold the uncompressed data.
;
;           MRCI_ERROR_BAD_MRC_FORMAT
;               The compressed data format was invalid (generally only
;               detectable as an overrun of the source buffer length, because
;               the MRC Format has no redundancy).
;               NOTE: Most software implementations will not generate this
;                     error, since it is to expensive (in time) to check for
;                     buffer overrun.  Hardware implementations may be able
;                     to check for this without performance penalty, however.
;
;       Uses
;           ax,flags

MRCDecompress proc near
;*  Save caller's registers
        SaveReg <bx,cx,dx,bp,si,di,ds>

;*  Enter a Windows disk critical section.  This must be done to prevent
;   the MCRI server from being reentered if multiple VMs are making MRCI
;   calls under 386 Enhanced mode Windows.  The MRCI server sets the
;   InDOS flag, but that is not enough, as InDOS is a per-VM variable.

        push    ax                      ; NOTE: Use this exact sequence of
        mov     ax,8001h                ;       instructions because Windows
        int     2ah                     ;       expects it and will patch
        pop     ax                      ;       other code here

        mov     ax,micapDECOMPRESS
        call    dword ptr es:[bx].mi_pfnOperate ; Call Server

;*  Release the disk critical section.  The MRCI server is now available.

        push    ax                      ; NOTE: Use this exact sequence of
        mov     ax,8101h                ;       instructions because Windows
        int     2ah                     ;       expects it and will patch
        pop     ax                      ;       other code here

;*  Restore caller's registers and exit
        RestoreReg <ds,di,si,bp,dx,cx,bx>
        ret
MRCDecompress endp

endm ;; DefineMRCDecompress
