;***    MRCSRV.ASM - MRCI Server EXAMPLE
;
;       MRCI version 1.00.08  02-Apr-1993
;
;       NOTE: This example is implemented assuming the code segment is
;             writeable, for simplicity.
;
;       EXPORTED FUNCTIONS:
;           Server    - intMRCI interrupt handler
;           Operate   - Compress/Decompress entry point
;           Strategy  - DOS device driver strategy entry point
;           Interrupt - DOS device driver interrupt entry point
;
;
;       INTERNAL FUNCTIONS:
;           InitMRCServer       - Install Server
;           ShouldWeInstall     - Test if we should replace existing server
;           GetInDOSFlagPointer - Get address of InDOS flag
;           DoCompress          - Unimplemented compress routine
;           DoDecompress        - Unimplemented decompress routine


include utility.inc
include mrci.inc


;***********************
;* Private Definitions *
;***********************

MYVERSION   equ 0100h                   ; Version of this server
MRCIVERSION equ 0100h                   ; Version of MRCI supported by this server
MYVENDOR    equ 5446534Dh               ; "MSFT" (reversed for byte ordering)
MYVENDORhi  equ 5446h                   ; High word of MYVENDOR
MYVENDORlo  equ     534Dh               ; Low word of MYVENDOR
cbMAX       equ 8192                    ; Maximum compression block size

;* This server's capabilities
micapMine   equ micapSTANDARD or micapDECOMPRESS


;******************
;* PRIVATE MACROS *
;******************


;***    EnterCriticalSection - Grab server critical section
;
;       Entry
;           cx - Flag indicating whether the caller may already be inside DOS.
;                mcAPPLICATION (0) - InDOS must be 0.
;                mcSYSTEM (1)      - InDOS may have any value.
;
;       Exit-Success
;           Zero Flag Set, critical section entered
;           InDos is guaranteed to be non-zero
;           fBusy set to non-zero
;
;       Exit-Failure
;           Zero Flag clear, critical section NOT entered
;
;       Uses
;           dx,si,ds,flags

EnterCriticalSection macro  reg
        cli                             ; Grab our lock atomically

        cmp     cs:fBusy,0              ; Are we already working on something?
        jnz     ecsx                    ;   YES, return failure

;*  Make InDOS non-zero
;
;   NOTE: We check for failure here.  We should not have to, since if we
;         are busy, fBusy would be set, and we would have detected it above.
;         The other case is a non-file system client calling, but InDOS is
;         set.  This case should not happen, because if InDOS is set, no
;         non-file system client should gain control of the CPU and try to
;         call us.  So, if such a non-file system client tries this, we
;         fail the call.
;
        lds     si,cs:fpInDOS           ; ds:si -> InDOS flag
        mov     dl,ds:[si]              ; dl = InDOS flag
        or      dl,dl                   ; In DOS?
        jz      ecs10                   ;   NO, okay to take critical section

;*  InDOS is non-zero, see if caller allows this
;   Zero Flag is CLEAR (=> not zero) from falling through the "jz" above
;
;   NOTE: We allow a system client to call even if InDOS is not set, since
;         it is possible that a system client may need to operate at
;         interrupt time for "background" operations.
;
;         *** THIS IS DISCOURAGED BEHAVIOR ***
;
        jcxz    ecsx                    ;   NO, caller is not file system
.errnz mcAPPLICATION

;*  InDOS was zero, or caller was a file system

ecs10:  mov     cs:oldInDOS,dl          ; Save old InDOS flag, for later
        mov     byte ptr ds:[si],1      ; Set InDOS flag
        mov     cs:fBusy,1              ; We are now busy!
        xor     dl,dl                   ; Set zero flag to indicate success

ecsx:   ; done with macro, zero flag indicates success/failure

        sti                             ; Done being atomic
endm ;; EnterCriticalSection


;***    LeaveCriticalSection - Leave server critical section
;
;       Entry
;           cx - Flag indicating whether the caller may already be inside DOS.
;                mcAPPLICATION (0) - Application client
;                mcSYSTEM (1)      - File System client
;
;       Exit
;           InDos restored to value saved by EnterCriticalSection
;           fBusy set to zero
;
;       Uses
;           ax,si,ds,flags

LeaveCriticalSection macro  reg
        cli                             ; Release our lock atomically

;; Should assert that fBusy is set
;;      cmp     cs:fBusy,0              ; Are we already working on something?
;;      jnz     lcs20                   ;   YES, but not any longer
;;lcs10:
;;
;;      Handle case of server attempting to leave critical section
;;      without being inside critical section!
;;lcs20:

;*  Restore InDOS
        lds     si,cs:fpInDOS           ; ds:si -> InDOS flag

;; Should assert that InDOS is non-zero
;;      cmp     ds:[si],0               ; InDOS set?
;;      jz      lcs10                   ;   NO, but should have been!

        mov     al,cs:oldInDOS          ; al = old InDOS flag
        mov     ds:[si],al              ; Restore old InDOS flag

;*  Clear our busy flag
        mov     cs:fBusy,0              ; No longer busy

        sti                             ; Done being atomic
endm ;; LeaveCriticalSection


;****************
;* CODE segment *
;****************

code    segment public byte
    assume  cs:code,ds:nothing,es:nothing,ss:nothing

;;      org     0       ; Driver starts at 0

;*************
;* DATA area *
;*************

;**     Character Device Header
;

DHlink  dd      0FFFFFFFFh              ; Next device => none
DHattr  dw      8000h                   ; Simple character device
DHstrat dw      Strategy                ; Device strategy entry point
DHinter dw      Interrupt               ; Device interrupt entry point
DHname  db      'MRCISRV$'              ; Device name


;**     mrcinfo - MRCINFO to return to clients
;
;       NOTE: This structure must be writeable, so that a new Server
;             can install and hook this structure with its values.
;
;       vendor,server version,MRCI version,entry point,capabilities,cbMAX
;
mi      MRCINFO <MYVENDOR,MYVERSION,MRCIVERSION,Operate,micapMine,cbMAX>


;**     fpOldintMRCI - previous contents of intMRCI interrupt vector
;
fpOldintMRCI   dd  ?


;**     pmi - pointer to active MRCINFO structure
;
;       If we replace an old server, then that server's MRCINFO structure
;       is the one that we use.  So, we need a level of indirection to it.
;       The more common case is that we are the only server, so this
;       variable will just point to our mi.
;
pmi     dd  ?


;**     rpDOS - pointer to DOS request packet
;
rpDos   dd      ?


;**     fpInDOS, oldInDOS - pointer to DOS "InDOS" flag, old InDOS flag value
;
;       This is REQUIRED for a software-only MRCI server, to signal that
;       the server is busy, and so prevent reentrancy.  Whenever the server
;       is about to become non-reentrant (starting a compress/decompress
;       operation, for example), it must ensure that InDOS is set.
;
;   ATTENTION--BEGIN
;
;       Furthermore, ALL callers of MRCI *must* ensure that the Windows
;       Disk Critical Section is held *prior* to calling MRCI.  This prevents
;       programs in different Windows virtual machines from reentering MRCI
;       and encountering the MRCI_ERROR_BUSY error.
;
;       Block device drivers, and system extensions like DBLSPACE.BIN, do not
;       need to grab this critical section because the MS-DOS kernel has
;       already done so before giving them control.
;
;       But TSRs and application programs *must* grab this critical section
;       prior to calling MRCI.
;
;   ATTENTION--END
;
;       The following worst-case scenario demonstrates why this is needed,
;       and why returning *Busy* (see fBusy below) is of no help:
;
;       1)  MSBackup calls MRCI server to do a decompress operation
;       2)  While MRCI server is busy, user brings up a TSR (e.g., SideKick)
;       3)  SideKick does a DOS file open and read on a compressed drive
;       4)  DOS FAT file system calls DoubleSpace
;       5)  DoubleSpace calls MRCI server, which returns busy
;
;       At this point, the system is deadlocked!  DoubleSpace has no choice
;       but to try the request again, or return some sort of disk error!
;
;       By setting the DOS InDOS flag, however, the MRCI server signals to
;       TSRs (and Windows) that it is entering a critical section, and so
;       a well-written TSR will not try to make DOS calls or MRCI calls.

fpInDOS     dd      ?   ; Pointer to DOS InDOS flag
oldInDOS    db      ?   ; Old InDOS flag value


;**     fBusy - TRUE if we are compressing/decompressing
;
;       Setting the DOS InDOS flag will protect us in most cases, but
;       having this flag lets us prevent being reentered by an interrupt-
;       time caller.
;
fBusy   db      ?   ; 0 => not busy, !0 => busy


;********
;* CODE *
;********


;***    Strategy - Device driver strategy entry point
;
;       Entry
;           es:bx - pointer to DOS request packet
;
;       Exit
;           es:bx saved in data:rpDos
;
;       Uses
;           None
;
Strategy    proc    far
    assume  cs:code,ds:nothing,es:nothing
        mov     cs:rpDos.offst,bx       ; Store request packet pointer
        mov     cs:rpDos.segmt,es
        ret
Strategy    endp


;***    Interrupt - Device driver interrupt entry point
;
;       Entry
;           rpDOS - pointer to DOS request packet saved by Strategy function
;
;       Exit
;           ???
;
;       Uses
;           None
;
Interrupt   proc    far
        SaveReg <ax,bx,cx,dx,si,di,ds,es>

        push    cs                      ; Make access to driver segment faster
        pop     ds
    assume  ds:code

        les     bx,rpDos                ; es:bx -> DOS request packet
        mov     al,es:[bx].rpFunction   ; al = function
        cmp     al,DEVINIT              ; Init call?
        mov     ax,STERR+03h            ; Assume ERROR+"Unknown command"
        jne     itrx                    ;   NO, fail call

;*  Init device

        call    GetInDOSFlagPointer     ; Needed for critical section calls

        SaveReg <bx,es>
        call    InitMRCServer
        RestoreReg <es,bx>
        mov     cx,OFFSET cs:EndOfResidentCode ; Assume we are installed
        or      ax,ax                   ; Are we installed?
        jz      itr10                   ;   YES, set break address, note that
                                        ;   ax=status=fine!

;*  Did not install MRCI server
        xor     cx,cx                   ;   NO, make break address small
        mov     ax,STERR                ; What error code should we use?

;*  Set break address
;
;   cx = break address
;   ax = status code

itr10:  mov     es:[bx+14].offst,cx
        mov     es:[bx+14].segmt,cs

;*  es:bx -> DOS request packet
;   ax = status code

itrx:   or      ax,STDONE               ; Set DONE bit in status
        mov es:[bx].rpStatus,ax         ; Store status
        RestoreReg <es,ds,di,si,dx,cx,bx,ax>
        ret
Interrupt   endp


;***    Server - MRCI server interrupt entry point
;
;       Entry
;           mov ax,MRCI function  ; Set function
;           int intMRCI           ; Call us here
;
;       Exit
;           See MRCI spec.  IRET to return.
;
;       Uses
;           AX, DI, ES, flags

Server  proc
    assume  ds:nothing,es:nothing,ss:nothing

;*  Verify that caller is asking for an MRCI server

        cmp     cx,sigOLD_CX            ; Signature match?
        je      srv10                   ;   YES, keep testing

        jmp     dword ptr cs:[fpOldintMRCI] ; Chain to previous hooker

srv10:  cmp     dx,sigOLD_DX            ; Signature match?
        je      srv20                   ;   YES, caller wants us

        jmp     dword ptr cs:[fpOldintMRCI] ; Chain to previous hooker

;*  Dispatch MRCI function

srv20:  cmp     ax,mrciDETECT           ; Detect call?
        je      dms10                   ;   NO, continue

;*  mrciQUERY call

        push    cs
        pop     es
        mov     di,OFFSET mi            ; es:di -> our MRCINFO structure

;*  Update CX/DX signatures so caller knows we are an MRCI server

                               ;   entry: cx='ab'  dx='cd'
        xchg    ch,cl                   ; cx='ba'  dx='cd'
        xchg    dh,dl                   ; cx='ba'  dx='dc'
        xchg    dx,cx                   ; cx='dc'  dx='ba'

        xor     ax,ax                   ; Indicate success
        iret                            ; Return to caller

dms10:  mov     ax,1                    ; Indicate failure

;*  EmptyIRET - address for fpOldintMRCI if intMRCI was unhooked.
;
;   To improve performance of Server when it chains a non-intMRCI request,
;   if we are the first hooker of intMRCI, we store the pointer to this
;   IRET.  This removes the need for Server to check that the fpOldintMRCI
;   value is valid, to prevent from INTing into hyperspace.
;
EmptyIRET:
        iret
Server  endp


;***    Operate - MRCI server operation entry point
;
;       Handle Compress/Decompress requests.
;
;       Entry
;              ax = Requested operation (micapXXXX)
;              cx = File System Caller flag
;                       mcAPPLICATION (0) - Application client
;                       mcSYSTEM (1)      - File System client
;           ds:si -> MRCREQUEST packet
;
;       Exit-Success
;           ax = 0
;           See MRCI spec for details on returned values.
;
;       Exit-Failure
;           See MRCI spec for details on returned values.
;
;       Uses
;           All (but CS:IP, SS:SP)

Operate proc    far
    assume  ds:nothing,es:nothing,ss:nothing

        and     ax,micapMine            ; Is this an operation we support?
        jz      opre                    ;   NO, fail request

;*  Make sure we are not being reentered

        EnterCriticalSection cx         ; Grab critical section
        jc      opre1                   ;   Already grabbed, go fail

        SaveReg <cx>                    ; Save fFileSystemCaller flag

;*  Dispatch operation
;
;   NOTE: We assume Standard Compress is the most common operation,
;         so we optimize the flow to reduce jumps for that operation.
;
        cmp     ax,micapSTANDARD        ; Standard compress?
        jz      opr10                   ;   YES, go do it

        cmp     ax,micapDECOMPRESS      ; Decompress?
        jnz     opre                    ;   NO, invalid operation (user had
                                        ;   more than one bit set!)
;*  Do Decompress

opr20:  call    DoDecompress
        jmp     short oprLCS

;*  Do Standard Compress

opr10:  call    DoCompress

;*  Release critical section
oprLCS: RestoreReg <cx>                 ; Get fFileSystemCaller flag
        LeaveCriticalSection cx         ; Leave critical section

        ret                             ; Return status


opre:   mov     ax,MRCI_ERROR_NOT_SUPPORTED
        ret

opre1:  mov     ax,MRCI_ERROR_BUSY
        ret
Operate endp


;***    DoCompress - Unimplemented compress routine
;
;
DoCompress  proc    near
DoCompress  endp


;***    DoDecompress - Unimplemented decompress routine
;
;
DoDecompress  proc    near
DoDecompress  endp


;******************
;* INIT-TIME CODE *
;******************

EndOfResidentCode  label  byte  ; Truncate device here after init is complete


;** BEGIN: Include code for MRCQuery here!
;
        DefineMRCQuery
;
;** END:   Include code for MRCQuery here!


;**     InitMRCServer - Initialize our MRCI Server
;
;       If an MRCI server IS NOT present, then we install our server.
;
;       If an MRCI server IS present, then we get the pointer to its
;       MRCINFO structure, and check to see if we are "better" than the
;       present server.  If so, then we install, and edit the old
;       server's MRCINFO structure so that its clients can call us.
;
;       Entry
;           none.
;
;       Exit-Success
;           ax = 0
;           Server installed
;           pmi = pointer to MRCI server MRCINFO structure
;
;       Exit-Failure
;           ax = 1, another, better server already present
;
;       Uses
;           ax,bx,si,ds,es,flags

InitMRCServer   proc    near
    assume  ds:nothing,es:nothing,ss:nothing

        call    MRCQuery                ; Check for MRCI server presence
        or      ax,ax                   ; Is server present?
        jz      ims20                   ;   NO, install ourselves

;*  Existing Server present
;
;   es:di -> old server's MCRINFO structure
;
;   We need to copy our MRCINFO structure over the old server's MCRINFO
;   structure, so that clients of the previous server that call the
;   server indirectly using the old MRCINFO structure will call us,
;   instead.
;
;   NOTE: We are guaranteed that no one is using the old server right now,
;         because we have the CPU.
;
;         So, we disable interrupts while we update the old MRCINFO structure.

        ;; dbgPrint 'Existing server present'

        call    ShouldWeInstall
        or      ax,ax                   ; Do install?
        jz      ims10                   ;   YES, go install

        ret                             ;   NO, return "did not install"

;*  Copy our MRCINFO to old server's MRCINFO

ims10:  cli                             ; Protect old MRCINFO from being
                                        ; in an inconsistent state.
        push    cs
        pop     ds
    assume  ds:code

        mov     si,OFFSET mi            ; ds:si -> our MRCINFO structure
        mov     cx,SIZE mi
        cld
        rep     movsb                   ; Copy our MRCINFO to old one

;*  Fall through to hook interrupt chain.
;
;   NOTE: Interrupts remain off, until we have hooked the interrupt chain.
;         We have to leave the old server in the chain, so that it can
;         chain on to previous non-MRCI hookers.

;*  Hook into intMRCI chain

ims20:  xor     ax,ax                   ; Segment of interrupt vector table
        mov     ds,ax
    assume  ds:nothing
        mov     si,intMRCI*4            ; ds:si -> intMRCI vector

        cli                             ; Editing the interrupt vector table

        mov     ax,OFFSET Server
        xchg    ds:[si].offst,ax        ; Set offset, ax = old offset
        mov     bx,cs
        xchg    ds:[si].segmt,bx        ; Store segment, bx = old segment

        or      ax,ax                   ; Old pointer valid?
        jnz     ims30                   ;   YES, go save

        mov     ax,OFFSET EmptyIRET     ;   NO, point at on of our IRETs
        mov     bx,cx

ims30:  mov     cs:fpOldintMRCI.offst,ax ; Save old offset, for chaining
        mov     cs:fpOldintMRCI.segmt,bx ; Save old segment, for chaining

        sti                             ; Done editing the intvec table

        ret
InitMRCServer   endp


;***    ShouldWeInstall - check if our server should supercede another server
;
;       We can install only if all of the following are true:
;           1) Our capabilities are at least as good as present server
;           2) Our MRCI version is at least as high
;           3) If the vendor is the same, our vendor server version must
;              be greater than the present server.  If the vendor does not
;              match, then we cannot compare vendor versions, so we assume
;              we should install.
;
;       Entry
;           es:di -> MRCINFO structure of current installed server
;
;       Exit-Success
;           ax = 0, we should install
;
;       Exit-Failure
;           ax = 1, we should not install
;           Existing driver is more capable than we are.
;
;       Uses
;           ax,bx,flags

ShouldWeInstall proc    near
;*  Our capabilities must be at least as good as present server

        mov     ax,es:[di].mi_flCapability ; server capabilities
        and     ax,not micapMine        ; ax = server caps that I do not have
        or      ax,ax                   ; Does the server have caps I do not?
        jnz     swie                    ;   YES, do NOT install

;*  We are at least as good on capabilities, so...
;*  Compare MRCI versions

        mov     ax,es:[di].mi_wMRCIVersion ; ax = present server MRCI version
        cmp     ax,MRCIVERSION          ; Are we at least as good as server?
        ja      swie                    ;   NO, do not install
        jb      swi10                   ; We are better than server, install

;*  We support the same MRCI version, so...
;*  See if we supercede the old <vendor,vendor version>

        mov     ax,es:[di].mi_lVendor.loword
        cmp     ax,MYVENDORlo           ; Match low word?
        jne     swi10                   ;   NO, DO install

        mov     ax,es:[di].mi_lVendor.hiword
        cmp     ax,MYVENDORhi           ; Match high word?
        jne     swi10                   ;   NO, DO install

        mov     ax,es:[di].mi_wVendorVersion
        cmp     ax,MYVERSION            ; Am I a newer driver version?
        jae     swie                    ;   NO, do NOT install

;*  Tests pass -- we should install

swi10:  xor     ax,ax                   ; Indicate DO install
        ret

swie:   mov     ax,1                    ; Indicate do NOT install
        ret

ShouldWeInstall endp


;***    GetInDOSFlagPointer - Get address of InDOS flag
;
;       Entry
;           none
;
;       Exit
;           fpInDOS = pointer to DOS InDOS flag
;
;       Uses
;           ax,flags

GetInDOSFlagPointer proc    near
        SaveReg <bx,es>
        mov     ah,34h
        int     21h                     ; Get In DOS flag

        mov     cs:fpInDOS.offst,bx
        mov     cs:fpInDOS.segmt,es
        RestoreReg <es,bx>
        ret

GetInDOSFlagPointer endp

code    ends

        end
